"""分布锁管理。
"""

import logging
from zenutils import importutils
from .exceptions import GlobalLockClassNotFoundException
from .config import Config

_logger = logging.getLogger(__name__)


class GlobalLockImplementationBase:
    """锁实现基础类。"""

    def __init__(
        self, config, name, timeout, sleep, blocking, blocking_timeout, **other_options
    ):
        # create properties
        self.name = name
        self.timeout = timeout
        self.sleep = sleep
        self.blocking = blocking
        self.blocking_timeout = blocking_timeout
        self.other_options = other_options
        self.lock = None
        self._is_locked = False
        # load config items
        self.config = config

    @property
    def lock_key(self):
        """计算出锁在引擎中的key值。"""
        return self.config.global_lock_key_prefix + self.name

    @property
    def is_locked(self):
        """当前上下文是否已经获得锁。"""
        return self._is_locked

    def __enter__(self, *args, **kwargs):
        """进入锁定上下文。"""
        self._is_locked = self.acquire()
        return self

    def __exit__(self, *args, **kwargs):
        """离开锁定上下文。"""
        self.release()

    def acquire(self) -> bool:
        """请求锁。返回锁定结果。

        @result: bool, True表示锁定成功，False表示锁定失败。
        """
        raise NotImplementedError()

    def release(self):
        """释放锁。"""
        if self._is_locked and self.lock:
            try:
                self.lock.release()
            except Exception as error:  # pylint: disable=broad-exception-caught
                _logger.warning(
                    "release lock failed, mostly becuase of the main process takes more time than the lock's expire time or the lock engine connection failed: %s...",
                    error,
                )
            self._is_locked = False


class GlobalLockManager:
    """分布锁管理器。"""

    def __init__(self, config):
        self.config = Config(config or {})
        # pylint: disable=no-member
        self.lock_class = importutils.import_from_string(
            self.config.global_lock_engine_class
        )
        if self.lock_class is None:
            raise GlobalLockClassNotFoundException(
                f"global lock class {self.config.global_lock_engine_class} is NOT found..."
            )

    def lock(
        self,
        name,
        timeout=None,
        sleep=None,
        blocking=None,
        blocking_timeout=None,
    ) -> GlobalLockImplementationBase:
        """获取锁。使用with语法：

        ```
        with lockman.lock(lock_name, timeout=5, blocking=True, blocking_timeout=1) as lock:
            if lock.is_locked:
                pass # do something with lock
            else:
                pass # do something without lock, mostly does NOTHING
        ```
        """
        if timeout is None:
            timeout = self.config.get("timeout", None)
        if sleep is None:
            sleep = self.config.get("sleep", 0.1)
        if blocking is None:
            blocking = self.config.get("blocking", True)
        if blocking_timeout is None:
            blocking_timeout = self.config.get("blocking_timeout", None)
        return self.lock_class(
            self.config,
            name,
            timeout=timeout,
            sleep=sleep,
            blocking=blocking,
            blocking_timeout=blocking_timeout,
        )



"""常量及关联配置项。
"""

# 支持的分布式锁实现
REDIS_GLOBAL_LOCK_CLASS = "globallock.redis_global_lock.RedisGlobalLock"
DJANGO_REDIS_GLOBAL_LOCK_CLASS = (
    "globallock.django_redis_global_lock.DjangoRedisGlobalLock"
)
ETCD_GLOBAL_LOCK_CLASS = "globallock.etcd_global_lock.EtcdGlobalLock"
ZOOKEEPER_GLOBAL_LOCK_CLASS = "globallock.zookeeper_global_lock.ZookeeperGlobalLock"
# global_lock_engine_class
DEFAULT_GLOBAL_LOCK_ENGINE_CLASS = REDIS_GLOBAL_LOCK_CLASS
# global_lock_engine_options
DEFAULT_GLOBAL_LOCK_ENGINE_OPTIONS = {}
# global_lock_key_prefix
DEFAULT_GLOBAL_LOCK_KEY_PREFIX = "_glocks:"



"""基于redis实现的分布式锁定。
"""

from kazoo.client import KazooClient
from zenutils import cacheutils
from .base import GlobalLockImplementationBase


class ZookeeperGlobalLock(GlobalLockImplementationBase):
    """基于zookeeper实现的分布式锁定。"""

    def __init__(self, config, name, **options):
        super().__init__(config, name, **options)
        self.lock = self.get_lock()

    @property
    def lock_key(self):
        lkey = self.config.global_lock_key_prefix.replace(":", "/") + self.name
        if not lkey.startswith("/"):
            lkey = "/" + lkey
        return lkey

    def acquire(self) -> bool:
        """请求锁。返回锁定结果。

        @result: bool, True表示锁定成功，False表示锁定失败。
        """
        return self.lock.acquire(
            blocking=self.blocking,
            timeout=self.blocking_timeout,
        )

    def get_lock(self):
        """获得锁对象。"""
        return self.get_zk_client().Lock(self.lock_key)

    @cacheutils.simple_cache
    def get_zk_client(self):
        """根据配置，生成连接池。

        所有锁管理器均共享这个连接池。
        """
        zookeeper_client = KazooClient(**self.config.global_lock_engine_options)
        zookeeper_client.start()
        return zookeeper_client



"""关于EtcdGlobalLock的单元测试。
"""
import time
from unittest import TestCase

from globallock.base import GlobalLockManager
import test_etcd_global_lock_local_settings


class TestEtcdGlobalLock(TestCase):
    """EtcdGlobalLock测试用例。"""

    def test1(self):
        """测试EtcdGlobalLock是否可用。"""
        lockman = GlobalLockManager(test_etcd_global_lock_local_settings.config)
        lockname = "test test test"
        with lockman.lock(
            lockname,
            timeout=5,
            blocking_timeout=1,
        ) as lock:
            if lock.is_locked:
                assert True
            else:
                assert False

    def test2(self):
        """测试timeout的有效性。"""
        lockman = GlobalLockManager(test_etcd_global_lock_local_settings.config)
        lockname = "test/test/test"
        lock = lockman.lock(
            lockname,
            timeout=1,
            blocking_timeout=1,
        )
        assert lock.acquire()
        assert lock.acquire() is False
        time.sleep(3)
        assert lock.acquire()  # etcd锁超时f起作用
        lock.release()



from .base import GlobalLockManager
from .base import GlobalLockImplementationBase
from .constants import REDIS_GLOBAL_LOCK_CLASS
from .constants import DJANGO_REDIS_GLOBAL_LOCK_CLASS
from .constants import ETCD_GLOBAL_LOCK_CLASS
from .constants import ZOOKEEPER_GLOBAL_LOCK_CLASS
from .config import Config



"""基于redis实现的分布式锁定。
"""

import redis
from zenutils import cacheutils
from .base import GlobalLockImplementationBase


class RedisGlobalLock(GlobalLockImplementationBase):
    """基于redis实现的分布式锁定。"""

    def __init__(self, config, name, **options):
        super().__init__(config, name, **options)
        self.lock = self.get_lock()

    def acquire(self) -> bool:
        """请求锁。返回锁定结果。

        @result: bool, True表示锁定成功，False表示锁定失败。
        """
        return self.lock.acquire()

    def get_lock(self):
        """获得锁对象。"""
        return redis.Redis(connection_pool=self.get_redis_connection_pool()).lock(
            self.lock_key,
            timeout=self.timeout,
            sleep=self.sleep,
            blocking=self.blocking,
            blocking_timeout=self.blocking_timeout,
            **self.other_options
        )

    @cacheutils.simple_cache
    def get_redis_connection_pool(self):
        """根据配置，生成连接池。

        所有锁管理器均共享这个连接池。
        """
        return redis.ConnectionPool(**self.config.global_lock_engine_options)



"""配置类。
"""

from zenutils import dictutils
from .constants import DEFAULT_GLOBAL_LOCK_ENGINE_CLASS
from .constants import DEFAULT_GLOBAL_LOCK_ENGINE_OPTIONS
from .constants import DEFAULT_GLOBAL_LOCK_KEY_PREFIX


class Config(dictutils.Object):
    """配置类。

    @todo: 后续统一修改为支持服务端统一管理、客户端缓存、实时更新机制的配置管理。
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.setdefault(
            "global_lock_engine_class",
            DEFAULT_GLOBAL_LOCK_ENGINE_CLASS,
        )
        self.setdefault(
            "global_lock_engine_options",
            DEFAULT_GLOBAL_LOCK_ENGINE_OPTIONS,
        )
        self.setdefault(
            "global_lock_key_prefix",
            DEFAULT_GLOBAL_LOCK_KEY_PREFIX,
        )



import copy
from django.conf import settings
from .base import GlobalLockManager
from .constants import DJANGO_REDIS_GLOBAL_LOCK_CLASS

# 从settings.py中获取全局分布式锁的设置
GLOBAL_LOCK_CONFIG = getattr(
    settings,
    "GLOBAL_LOCK_CONFIG",
    {},
)


def _get_config():
    config = copy.deepcopy(GLOBAL_LOCK_CONFIG)
    if not "global_lock_engine_class" in config:
        config["global_lock_engine_class"] = DJANGO_REDIS_GLOBAL_LOCK_CLASS
    if not "global_lock_engine_options" in config:
        config["global_lock_engine_options"] = {}
    if not "redis-cache-name" in config["global_lock_engine_options"]:
        config["global_lock_engine_options"]["redis-cache-name"] = "default"
    return config


def get_default_global_lock_manager():
    """根据Django settings.py的设置，获取分布式锁。"""
    config = _get_config()
    return GlobalLockManager(config)



"""基于etcd实现的分布式锁定。
"""

import etcd3
from .base import GlobalLockImplementationBase


class EtcdGlobalLock(GlobalLockImplementationBase):
    """基于etcd实现的分布式锁定。"""

    def __init__(self, config, name, **options):
        super().__init__(config, name, **options)
        self.etcd_client = self.get_etcd3_client()
        self.lock = self.etcd_client.lock(
            self.lock_key,
            ttl=self.timeout,
        )

    @property
    def lock_key(self):
        """
        etcd3中Lock自带/locks/前缀。
        所以我们自己的前缀不能再以/开头。
        """
        lkey = self.config.global_lock_key_prefix.replace(":", "/") + self.name
        lkey = lkey.lstrip("/")
        return lkey

    def acquire(self) -> bool:
        """
        请求锁。返回锁定结果。

        @result: bool, True表示锁定成功，False表示锁定失败。
        """
        if self.blocking:
            blocking_timeout = self.blocking_timeout
        else:
            blocking_timeout = 0
        return self.lock.acquire(timeout=blocking_timeout)

    def get_etcd3_client(self):
        """
        根据配置，生成连接池。

        所有锁管理器均共享这个连接池。
        """
        return etcd3.client(**self.config.global_lock_engine_options)



"""关于RedisGlobalLock的单元测试。
"""
import uuid
from unittest import TestCase

from globallock.base import GlobalLockManager
from globallock.exceptions import GlobalLockClassNotFoundException
import test_redis_global_lock_local_settings


class TestRedisGlobalLock(TestCase):
    """关于RedisGlobalLock的单元测试。"""

    def test1(self):
        """测试RedisGlobalLock是否可用。"""
        lockman = GlobalLockManager(test_redis_global_lock_local_settings.config)
        lockname = str(uuid.uuid4())
        with lockman.lock(
            lockname,
            timeout=5,
            blocking_timeout=1,
        ) as lock:
            if lock.is_locked:
                assert True
            else:
                assert False

    def test2(self):
        """测试RedisGlobalLock不可重入。"""
        lockman = GlobalLockManager(test_redis_global_lock_local_settings.config)
        lockname = str(uuid.uuid4())
        with lockman.lock(
            lockname,
            timeout=5,
            blocking=False,
        ) as lock:
            if lock.is_locked:
                assert True
                assert lock.acquire() is False
            else:
                assert False

    def test3(self):
        """测试实现类不存在的情况。"""
        config = {
            "global_lock_engine_class": "not_implemented_global_lock_class_path",
        }

        with self.assertRaises(GlobalLockClassNotFoundException):
            GlobalLockManager(config)



"""基于django-redis实现的分布式锁定。
"""

from django_redis import get_redis_connection
from .redis_global_lock import RedisGlobalLock


class DjangoRedisGlobalLock(RedisGlobalLock):
    """
    基于django-redis实现的分布式锁定。
    底层使用redis提供的分布式锁实现。
    与RedisGlobalLock的区别主要在于redis连接的获取方式，
    所以只需要重载get_lock函数即可。
    """

    def get_lock(self):
        """获得锁对象。"""
        return self.get_redis_connection().lock(
            self.lock_key,
            timeout=self.timeout,
            sleep=self.sleep,
            blocking=self.blocking,
            blocking_timeout=self.blocking_timeout,
            **self.other_options
        )

    def get_redis_connection(self):
        """从django的cache配置中获取redis连接。"""
        return get_redis_connection(
            self.config.global_lock_engine_options.get("redis-cache-name", "default")
        )
